"""
Copyright © 2025, UChicago Argonne, LLC

Full license found at _YOUR_INSTALLATION_DIRECTORY_/imgmarker/LICENSE
"""

from PyQt6.QtWidgets import (
    QApplication, QMainWindow, QPushButton, QLabel, 
    QScrollArea, QGraphicsView, QVBoxLayout, QWidget, 
    QHBoxLayout, QLineEdit, QInputDialog, QCheckBox, 
    QSlider, QLineEdit, QFileDialog, QFrame, 
    QDialog, QSizePolicy, QGraphicsPathItem, QGraphicsProxyWidget,
    QLineEdit, QGraphicsScene, QGraphicsPixmapItem, QSpinBox,
    QMessageBox, QTableWidget, QTableWidgetItem, QHeaderView,
    QMenu, QColorDialog, QWidgetAction,
)

from PyQt6.QtGui import (
    QIcon, QFont, QClipboard, QAction,
    QPen, QColor, QPixmap, QPainter, QPainterPathStroker,
    QPainterPath, QImage, QShortcut, QDesktopServices,
    QBrush
)
    
from PyQt6.QtCore import (
    Qt, QPoint, QKeyCombination, QPointF, 
    QEvent, QUrl, PYQT_VERSION_STR
)